#ifndef LIST_H
#define LIST_H
/**********************************************************************
 * The following classes implement a singly linked list with a head
 * node and a tail node. Some basic list functions are supported.
 **********************************************************************/
#include<iostream>
#include<iomanip>
#include<cstdlib> //for exit
#include<new>
#include <stdexcept>

using namespace std;//naughty


template< class T >
class List;// forward declaration, so Node can have List as a friend
template< class T >
class Node{
	friend class List<T>;
	friend class List<T>::iterator;
	friend class List<T>::const_iterator;
/* Only Lists should have easy access to Nodes, to protect the 
 * structural integrity of Lists. 
 */
	private:
		Node();// T must have a default constructor.
		Node(T, Node * n=NULL);// constructor wth elt
		T elt;
		Node *next;
};
template< class T >
Node< T >::Node()
{
	next=NULL;
}//constructor

template< class T >
Node< T >::Node(T val, Node *ptr): elt(val), next(ptr)
{
}//constructor

template <class T>
class List
{
	public:
		class iterator;//forward declaration
		class const_iterator;//forward declaration
		class iterator
		{
			friend class List<T>;
			friend class const_iterator;
			Node<T>* pointer;
			
			public:
			iterator():pointer(NULL){}
			iterator operator++()//prefix iterator advance
			{
				if (pointer->next!=NULL)
				{
					pointer=pointer->next;
					return *this;
				}
				else
					throw out_of_range(" cannot advance beyond end of list ");
			}
			iterator operator++(int)//postfix
			{
				iterator temp=*this;
				if (pointer->next!=NULL)
				{
					pointer=pointer->next;
					return temp;
				}
				else
					throw out_of_range(" cannot advance beyond end of list ");
			}
			T& operator*()//Return the elt of the Node 
			{
				return pointer->elt;
			}
			bool operator!=(iterator rhs)
			{
				return (pointer!=rhs.pointer);
			}
			
			
		};
			class const_iterator//for constant functions
		{
			friend class List<T>;
			friend class iterator;
			Node<T>* pointer;
			
			public:
			const_iterator():pointer(NULL){}
			const_iterator operator++()
			{
				if (pointer->next!=NULL)
				{
					pointer=pointer->next;
					return *this;
				}
				else
					throw out_of_range(" cannot advance beyond end of list ");
			}
			const_iterator operator++(int)//postfix
			{
				const_iterator temp=*this;
				if (pointer->next!=NULL)
				{
					pointer=pointer->next;
					return temp;
				}
				else
					throw out_of_range(" cannot advance beyond end of list ");
			}
			T operator*() 
			{
				return pointer->elt;
			}
			bool operator!=(const_iterator rhs)
			{
				return (pointer!=rhs.pointer);
			}
			bool operator!=(iterator rhs)
			{
				return (pointer!=rhs.pointer);
			}
			const_iterator(iterator rhs)//a conversion guide
			{
				pointer=rhs.pointer;
			}
			
		};
		List();
		~List();
		bool is_empty();
		void insert(iterator , T);//Insert x after the iterator
		void print();// Output list using an iterator
		bool delete_next(iterator);//Delete the element after the
					  //iterator. Return true if the
					  //delete is successful.
					  //Return false if the iterator
					  //was at the last element.
		void make_empty(); // Delete all Nodes except the head
				   // and tail.
		void push_front(T);//Insert at the beginning of the list
		void push_back(T);//Insert at the end of the list
		List(const List & rhs)// copy constructor
		{
			tail=new Node<T>;
			head=new Node<T>;
			head->next=tail;
			for(const_iterator it=rhs.begin();it!=rhs.end();it++)
				push_back(*it);
		}
			
		const List & operator=(const List & rhs);// assignment
	
		// return pointer to first Node after head Node	
		iterator begin();
		const_iterator begin() const;
		// return tail
		iterator end();
		const_iterator end() const;
		
	private:
		Node<T> *head;
		Node<T> *current_pos;
		Node<T> *tail;
		// return pointer to head
		iterator head_node();
};

		
	
template <class T>
List<T>::List()
{
	tail=new Node<T>;
	head=new Node<T>;
	head->next=tail;
}

template<class T>
	List<T>::~List()
{
	make_empty();// Only the memory for 'head' and tail
		     // is still allocated. 
	delete head;// Free that memory.
	delete tail;
	
}

template<class T>
const List<T>& List<T>::operator=(const List & rhs)
		{
			if(&rhs!=this)
			{
				make_empty();
				for(const_iterator it=rhs.begin();it!=rhs.end();it++)
					push_back(*it);
			
			}
			return *this;
		}	

template <class T>
bool List<T>::is_empty()
{
	if(head->next==tail)
		return true;
	else
		return false;
}
template <class T>
void List<T>::push_front(T x)
{
	Node<T> *p;
	p=new Node<T>(x, head->next);
	head->next=p;
}

template <class T>
void List<T>::push_back(T x)
{
	iterator it=head_node();
	while(it.pointer->next!=tail)
		it++;
	insert(it,x);
}

template <class T>
void List<T>::insert(iterator it, T x)// Insert x after the iterator
{
	Node<T> *p;
	p=new Node<T>(x,it.pointer->next);
	it.pointer->next=p;
}


template  <class T>
void List<T>::print()
{
	if(is_empty())
		cout<<"\nThe list is empty.";
	else
		for(iterator it=begin();it!=end();it++)
			cout<<*it<<" ";

}
	
template <class T>
bool List<T>::delete_next(iterator  it)
{
	if(it.pointer->next==tail)
		return false;
	else
	{
		Node<T> *p=it.pointer->next;
		it.pointer->next=(*p).next;
		delete p;
		return true;
	}
}

template <class T>
void List<T>::make_empty()
{
	iterator it=head_node();
	while(delete_next(it));
		
}

template <class T>
typename List<T>::iterator List<T>::begin()
		{
			iterator p;
			p.pointer=head->next;
			return p;
		}
template <class T>
typename List<T>::const_iterator List<T>::begin() const
		{
			iterator p;
			p.pointer=head->next;
			return p;
		}

template <class T>
typename List<T>::iterator List<T>::end()
		{
			iterator p;
			p.pointer=tail;
			return p;
		}
template <class T>
typename List<T>::const_iterator List<T>::end() const
		{
			const_iterator p;
			p.pointer=tail;
			return p;
		}


template <class T>
typename List<T>::iterator List<T>::head_node()
		{
			iterator p;
			p.pointer=head;
			return p;
		}
		
#endif
